/*
 * Copyright (C) 2003-2007 Funambol, Inc
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY, TITLE, NONINFRINGEMENT or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307  USA
 */

#include <stdlib.h>
#include "worker.h"
#include "s4jproxy.h"
#include "base/util/utils.h"
#include "notify/util.h"
#include "pim/maincpp.h"

#include "notify/addresschange.h"
#include "processutils.h"
#include "CTPManager.h"
#include "CTPParam.h"
#include <Pmpolicy.h>


// Worker thread
extern "C" DWORD WINAPI NotifWorker(LPVOID lpv);

static bool syncinprogress = false;

// Handle of the worker thread
static HANDLE handle = 0;
// Connection socket
static SOCKET sock = 0;
static S4JProxy* s4jp = 0;
// Message buffer
static char *message = 0;
static int msglen = 0;
// Thread start time
time_t startTime;

/**
* Concatenate a char buf with specified len to the message buffer
*/
static void addBuffer(const char *buf, int len);

/**
* Close the communication socket client / server. Remove gracefully
* all the buffers and classes involved in the listening process.
*/
static void closeWorker();

/**
* Return the message
*/
const char *getMsg() { return message; }

/**
* Return the message lenght
*/
int getLen() { return msglen; }


/**
* Get the current time as Unix time_t
*/
static time_t getTime()
{
    SYSTEMTIME st;
    FILETIME ft;

    GetSystemTime(&st);
    SystemTimeToFileTime(&st, &ft);
    // Convert FILETIME to time_t
    __int64 llTmp;
    memcpy (&llTmp, &ft, sizeof (__int64));
    llTmp = (llTmp - 116444736000000000) / 10000000;
    return (time_t) llTmp;
}


/**
* Create the thread on the socket client / server waiting for the server
* messages
*/
DWORD startWorker(SOCKET s)
{
    // Check worker
    if (handle) {
        LOG.debug("Another worker handle: %x\n", handle);
        DWORD exitcode = 0;
        if(GetExitCodeThread(handle, &exitcode)) {
            if (exitcode == STILL_ACTIVE) {
                LOG.debug("Worker thread still active");
                if (syncinprogress) {
                    LOG.debug("The sync is in progress");
                    LOG.debug("S4ND: connection busy");
                    return ERROR_BUSY;
                }
                else if( (startTime - getTime()) <  ASK_TIMEOUT ) {
                    LOG.debug("Request too close, wait");
                    LOG.debug("S4ND: connection busy");
                    return ERROR_BUSY;
                }
            }
        }
        LOG.debug("The sync is NOT in progress, closing thread");
        if(sock){
            closesocket(sock);
            sock= 0;
        }
        Sleep(100);
        if(handle){
            LOG.debug("Closing handle");
            CloseHandle(handle);
            handle=0;
        }
    }
    // Clean-up dirty messages
    if(message){
        delete [] message;
        message=0;
    }
    msglen=0;

    // Start worker
    sock = s;
    s4jp = S4JProxy::getInstance();

    startTime = getTime();   // record starting time
    handle = CreateThread(NULL, 0, NotifWorker, (LPVOID*) sock, 0, NULL);
    if (handle == 0) {
        LOG.debug("S4ND:ERROR:CreateThread fails. GetLastError() returns <0x%08x>", GetLastError());
        // Release resources
        closeWorker();
    }

    LOG.debug("S4ND: new worker thread handle=<0x%08x>", handle);
    return ERROR_SUCCESS;
}

/**
* Close the communication socket client / server
*/
DWORD stopWorker()
{
    if(sock){
        closesocket(sock);
        sock= 0;
    }
    Sleep(100);
    if ( !handle )
        return ERROR_SUCCESS;
    else
        return ERROR_BUSY;
}

static void closeWorker()
{
    if(sock){
        closesocket(sock);
        sock= 0;
    }
    if(handle){
        CloseHandle(handle);
        handle=0;
    }
    S4JProxy::dispose();
    if(message){
        delete [] message;
        message=0;
    }
    msglen=0;

    LOG.debug("Worker end.");
}

static void addBuffer(const char *buf, int len)
{
    if(!buf || !len)
        return;

    if(!message){
        message = new char[len+1];
        memcpy(message, buf, len);
        msglen = len;
    }
    else{
        char *tmp = new char[msglen+len+1];
        memcpy(tmp, message, msglen);
        memcpy(tmp+msglen, buf, len);
        msglen += len;
        delete [] message;
        message = tmp;
    }
}

/**
* Process the incoming server message and return a response.
* Return NULL if the message is not complete.
*/
const char *processPacket(const char *pkt, int pktlen)
{
    S4JProxy::RetCode code = S4JProxy::Error;
    static time_t last = getTime();

    time_t now = getTime();

    if ( (now - last) > 5 ){
        // reset message buffer
        delete [] message;
        message = 0;
    }

    // Message arrived
    LOG.debug("S4ND: %d bytes received", pktlen);

    addBuffer(pkt, pktlen);

    LOG.debug("S4ND: total message len %d", msglen);
    hexDump(message, msglen);

    code=s4jp->parsePkg0(message, msglen);
    LOG.debug("Return from parseMessage: %d", code);

    if (code != S4JProxy::Continue)
        return s4jp->getResponse();
    else
        return NULL;
}

/**
* Connection thread client/server. It waits for the server request,
* parses the message and if possible starts the synchronization of the
* server specified data source.
*/
extern "C" DWORD WINAPI NotifWorker(LPVOID lpv)
{
    SOCKET sock = (SOCKET)lpv;
    char buf[1501];
    int buflen=1500, pktlen;
    const char *response=NULL;

    while (!response) {
        pktlen = recv(sock, buf, buflen, 0);
        LOG.debug("Pktlen: %d", pktlen);
        switch(pktlen) {
            case SOCKET_ERROR: {
                DWORD err = WSAGetLastError();
                if(err == WSAEMSGSIZE ){
                    // TODO: this case should be handled
                    LOG.debug("Message larger than %d bytes!", buflen);
                    response = "500 - Message too long\n";
                }
                else {
                    LOG.error("Recv error: %d", err);
                    response = "500 - Network error\n"; // Try to respond anyway?
                }
                break;
            }
            case 0:
                // End of connection
                LOG.debug("Connection closed by peer.");
                response = "";  // Conn closed, no response
                break;
            default:
                response = processPacket(buf, pktlen);
        }
    }

    // If message is not empty, send it
    if(*response){
        LOG.debug("S4ND: Sending message <%s> to client", response);
        send(sock, response, strlen(response),0);
        // Start Sync
        if ( !s4jp->hasErrors() ) {
            closesocket(sock); // XXX
            syncinprogress=true;
            s4jp->sync();
            LOG.debug("Sync End.");
            syncinprogress=false;
        }
    }
    // Delete thread
    closeWorker();
    return 0;
}




/**
 * This is the main STP thread. Starts the notifyAddressChange sync
 * and returns the AN_ResponseCode: 0 if address accepted (success), 
 * otherwise > 0.
 */
DWORD WINAPI stpWorker(LPVOID lpv) {

    LOG.debug("Starting stpWorker thread");
    int ret = 0;

    AN_ResponseCode responseCode = notifyAddressChange(TEXT(APPLICATION_URI));
    if(responseCode == AN_AddressAccepted ){
        setServerNotified(true);
    }
    else {
        setServerNotified(false);
    }

    LOG.debug("Exiting stpWorker thread: response = %d", responseCode);
    return responseCode;
}



// -------------------- CTP --------------------------
/**
 * This is the main CTP thread. Manages the CTP connection using CTPManager:
 * 1. waits until the stpThread is finished, if successfully we exit this thread
 *    (STP is working), otherwise continue with CTP (STP is not working)
 * 2. Opens the CTP socket connection
 * 3. Authenticate the client
 * 4. Starts threads to receive messages and to send 'ready' msg.
 * 
 * @param lpv  the HANDLE of the stpThread running (NULL if not started)
 * @return a value >= 0 if no errors:
 *         0 if CTP closed correctly (Client closed CTP)
 *         1 if STP is working correctly, CTP not started
 *         2 if CTP client not authenticated (at the 2nd AUTH msg)
 *         3 if CTP client unauthorized
 *         4 if CTP client authentication forbidden
 *         <0 if errors
 *
 * @note This thread exits only in case of:
 *   - STPThread ended succesfully (STP is working, CTP not started)
 *   - CTP client not authenticated
 *   - CTP in 'leaving state' (Client is closing CTP)
 * In case of socket/Server errors, the whole CTP connection is pulled down
 * and restored again in a while (after 'ctpRetry' seconds).
 */
DWORD WINAPI ctpWorker(LPVOID lpv) {

    LOG.debug("Starting ctpWorker thread");
    int ret = 0;
    HANDLE stpThread = (HANDLE)lpv;

    // Get the unique instance of CTPManager.
    CTPManager* ctpManager = CTPManager::getInstance();

    //
    // Wait for STPThread to finish.
    // - if handle is NULL, no STP has been done  -> go with CTP.
    // - if STPThread returned error              -> go with CTP.
    // - if STPThread returned ok                 -> exit now (STP is running).
    //
    int timeout = ctpManager->getConfig()->getNotifyTimeout();
    if (stpThread) {
        LOG.debug("Waiting for STP notification sync to finish (timeout = %d sec)...", timeout);
        DWORD waitResult = WaitForSingleObject(stpThread, timeout*1000);
        switch (waitResult) {

            // Thread exited, get exit code.
            case WAIT_ABANDONED:
                LOG.debug("STPThread abandoned");
            case WAIT_OBJECT_0: {
                DWORD exitcode = 0;
                GetExitCodeThread(stpThread, &exitcode);
                LOG.debug("STPThread ended with code %d", exitcode);
                if (exitcode == 0) {
                    LOG.debug("STP is running correctly, close CTP if active");
                    CloseHandle(stpThread);
                    ctpManager->stopCTP();
                    return 1;
                }
                else {
                    // STP failed, go with CTP
                    LOG.debug("STP failed, go with CTP");
                    break;
                }
            }

            // Thread is still running after timeout -> kill it and continue.
            case WAIT_TIMEOUT: {
                LOG.debug("Timeout - STPThread will now be terminated");
                TerminateThread(stpThread, 1);
                break;
            }

            // Some error occurred (case WAIT_FAILED)
            default: {
                LOG.debug("Wait error on STPThread");
                break;
            }
        }
    }
    CloseHandle(stpThread);


    // Start the CTP connection process.
    // *********************************
    // Infinite cycle: always restore the connection if it's lost 
    // or in case of errors. Exit only if leavingState flag is up.
    //
    bool restore = false;
    bool jump    = false;
    while (ctpManager->isLeaving() == false) {

        if (restore) {
            // Restoring from a broken connection: close socket and wait some seconds.
            LOG.debug("Restoring CTP connection...");
            ctpManager->closeConnection();

            int sleepTime = ctpManager->getConfig()->getCtpRetry();
            sleepTime = min(sleepTime, ctpManager->getConfig()->getMaxCtpRetry());
            LOG.debug("Sleep for %d seconds...", sleepTime);
            Sleep(sleepTime * 1000);

            // Save the new value to config
            sleepTime *= CTP_RETRY_INCREASE_FACTOR;     // Double the retry time
            ctpManager->getConfig()->setCtpRetry(sleepTime);
            restore = false;
        }
        if (jump) {
            // Restoring from a JUMP status: close socket and reconnect immediately.
            LOG.debug("Restoring CTP connection from a JUMP...");
            ctpManager->closeConnection();
            jump = false;
        }

        //
        // Open socket connection
        // ----------------------
        LOG.debug("Open CTP connection...");
        if (ctpManager->openConnection()) {
            ctpManager->closeConnection();
            ret = -1;
            goto finally;
        }

        while (checkStartSync()) {
            // Wait for sync to finish: CTP AUTH phase uses the same clientNonce
            // used by sync authentication, so avoid interferences.
            LOG.debug("Waiting for sync to finish...");
            Sleep(5000);
        }

        //
        // Authentication
        // --------------
        LOG.debug("Sending [AUTH] message...");
        if (ctpManager->sendAuthMsg()) {
            ctpManager->closeConnection();
            ret = -2;
            goto finally;
        }
        
        // Receiving AUTH status message
        CTPMessage* authStatusMsg = ctpManager->receiveStatusMsg();
        if (!authStatusMsg) {
            LOG.error("Error receiving AUTH status msg.");
            restore = true;
            continue;
        }
        char authStatus = authStatusMsg->getGenericCommand();
        list<CTPParam>::iterator param;
        char* buf = NULL;
        switch (authStatus) {

            case ST_NOT_AUTHENTICATED:
                //
                // Retry with new nonce received
                //
                LOG.info("Client not authenticated: retry with new nonce");
                buf = extractNonceParam(authStatusMsg);
                if (buf) {
                    // Save new nonce to config, and save config to registry!
                    ctpManager->getConfig()->setCtpNonce(buf);
                    ctpManager->getConfig()->saveCTPConfig();
                    delete [] buf;
                }
                else {
                    // NONCE not found -> restore connection
                    LOG.error("Error receiving NON_AUTHENTICATED Status message: NONCE param is missing");
                    restore = true;
                    continue;
                }

                // Send 2nd auth msg
                LOG.info("Sending CTP authentication message...");
                if (ctpManager->sendAuthMsg()) {
                    ret = -3;
                    goto finally;
                }

                // Check 2nd status received, only OK allowed
                authStatusMsg = ctpManager->receiveStatusMsg();
                authStatus = authStatusMsg->getGenericCommand();
                if (authStatus == ST_OK) {
                    // *** Authentication OK! *** 
                    // Save nonce if any (go to case ST_OK)
                }
                else {
                    LOG.info("CTP error: Client not authenticated. Please check your credentials.");
                    ret = 2;
                    goto finally;
                }
                // no 'break': need to enter into case ST_OK...


            case ST_OK:
                // *** Authentication OK! *** 
                LOG.info("Client authenticated successfully!");
                ctpManager->setCtpState(CTP_STATE_READY);
                // Save nonce if any
                buf = extractNonceParam(authStatusMsg);
                if (buf) {
                    // Save new nonce to config, and save config to registry!
                    ctpManager->getConfig()->setCtpNonce(buf);
                    ctpManager->getConfig()->saveCTPConfig();
                    delete [] buf;
                }
                else {
                    LOG.info("No new nonce received.");
                }
                break;


            // --- note: JUMP not implemented Server side ----
            case ST_JUMP:
                //
                // Jump to desired server 'to' and save the 'from' value
                //
                LOG.info("Server requested a JUMP");
                if (authStatusMsg->params.size() < 1) {
                    // Expected FROM and TO params -> restore connection
                    LOG.error("Error receiving JUMP Status message: some parameter is missing");
                    restore = true;
                    continue;
                }

                // Read FROM and TO parameters and update CTPConfig
                param = authStatusMsg->params.begin();
                while (param != authStatusMsg->params.end()) {
                    int valueLen = (*param).getValueLength();
                    void* value  = (*param).getValue();
                    if ((*param).getParamCode() == P_FROM) {
                        char* from = stringdup((char*)value, valueLen);
                        ctpManager->getConfig()->setUrlFrom(from);
                        delete [] from;
                    }
                    else if ((*param).getParamCode() == P_TO) {
                        char* url = stringdup((char*)value, valueLen);
                        string to = ctpManager->getConfig()->getHostName(url);
                        int port  = ctpManager->getConfig()->getHostPort(url);
                        ctpManager->getConfig()->setUrlTo(to);
                        ctpManager->getConfig()->setCtpPort(port);
                        delete [] url;
                    }
                    else {
                        // Unexpected Status param -> restore connection
                        LOG.error("Error receiving JUMP Status message: unexpected param '0x%02x'", (*param).getParamCode());
                        restore = true;
                        continue;
                    }
                    param++;
                }

                // Now restore the socket connection to the new Server address...
                LOG.debug("JUMP status received: FROM %s TO %s:%d", ctpManager->getConfig()->getUrlFrom().c_str(), 
                                                                    ctpManager->getConfig()->getUrlTo().c_str(), 
                                                                    ctpManager->getConfig()->getCtpPort() );
                jump = true;
                continue;


            case ST_UNAUTHORIZED:
                // Not authorized -> save nonce if any, exit thread
                LOG.info("Unauthorized by the Server, please check your credentials.");
                buf = extractNonceParam(authStatusMsg);
                if (buf) {
                    // Save new nonce to config, and save config to registry!
                    ctpManager->getConfig()->setCtpNonce(buf);
                    ctpManager->getConfig()->saveCTPConfig();
                    delete [] buf;
                }
                else {
                    LOG.debug("No new nonce received.");
                }
                ret = 3;
                goto finally;


            case ST_FORBIDDEN:
                // Authentication forbidden -> exit thread
                LOG.info("Authentication forbidden by the Server, please check your credentials.");
                ret = 4;
                goto finally;

            case ST_ERROR:
                // Error -> restore connection
                LOG.info("Received ERROR status from Server: restore ctp connection");
                restore = true;
                continue;

            default:
                // Unexpected status -> restore connection
                LOG.error("Unexpected status received '0x%02x' -> restore ctp connection", authStatus);
                restore = true;
                continue;
        }

        //
        // Creates the thread that will be blocked waiting for server msg
        // and waits until it ends (errors or ctpConnTimeout).
        //
        ctpManager->receive();

        // If here, connection was broken or ctpConnTimeout -> restore connection
        restore = true;
    }

finally:

    LOG.debug("Exiting ctpWorker thread");
    return ret;
}



/**
 * This is used as a thread for the CTP process, called once CTP
 * is connected and ready to receive notifications from the Server.
 * We listen to the CTP socket connection, calling ctpManager->receiveStatusMsg().
 * - In case of error/wrong status from the Server, we get back to receiving.
 * - If a OK message is received, we get back to receiving.
 * - If a SYNC message is received, we start the sync process.
 *
 * @param lpv  unused
 * @note  exit only in case of errors or if we're in leaving state.
 * @return 0 if no errors
 */
DWORD WINAPI receiveWorker(LPVOID lpv) {

    LOG.debug("Starting receiveWorker thread");
    int ret = 0;
    CTPManager* ctpManager = CTPManager::getInstance();


    // Keep the socket open, always in 'receive' state.
    // Exit only in case of errors or if we're leaving CTP.
    while (ctpManager->isLeaving() == false) {

        // Receive msg from Server
        CTPMessage* statusMsg = ctpManager->receiveStatusMsg();
        if (!statusMsg) {
            // Error on receiving -> exit thread
            ret = -1;
            goto finally;
        }

        char status = statusMsg->getGenericCommand();
        SyncNotification* sn = NULL;
        switch (status) {

            case ST_OK:
                // 'OK' to our 'READY' command -> back to recv
                LOG.info("[OK] received -> back to receive state");
                break;

            case ST_SYNC:
                //
                // Start the sync!
                // ---------------
                LOG.info("[SYNC] notification received! Starting the sync");
                sn = statusMsg->getSyncNotification();
                startSyncFromSAN(sn);

                // Back to recv
                LOG.debug("Back to receive state");
                break;

            case ST_ERROR:
                LOG.debug("[ERROR] message received");
            default:
                // Error from server -> exit thread (will try restoring the socket from scratch)
                LOG.debug("Bad status received (code 0x%02x), exiting thread", status);
                ret = -2;
                goto finally;
        }
    }

finally:
    LOG.debug("Exiting receiveWorker thread");
    return ret;
}



/**
 * Thread used to send 'READY' messages as a heartbeat, every 'ctpReady' seconds.
 * It ends only if CTP is in 'leaving state' (Client stops CTP).
 * @param lpv  unused
 * @return 0 if no errors
 */
DWORD WINAPI heartbeatWorker(LPVOID lpv) {
    LOG.debug("Starting heartbeatWorker thread");
    int ret = 0;

    // Load the sleep interval (ctpReady)
    CTPManager* ctpManager = CTPManager::getInstance();
    int sleepInterval = ctpManager->getConfig()->getCtpReady();

    // To ensure the heartbeat is sent even if device is in suspended mode
    PowerPolicyNotify(PPN_UNATTENDEDMODE, TRUE);

    // Send 'ready' message to Server and sleep ctpReady seconds
    while (ctpManager->isLeaving() == false) {
        if (ctpManager->sendReadyMsg()) {
            LOG.debug("Error sending READY msg");
            ret = 1;
            break;
        }
        LOG.debug("Next ready msg will be sent in %d seconds...", sleepInterval);
        Sleep(sleepInterval * 1000);
    }

    PowerPolicyNotify(PPN_UNATTENDEDMODE, FALSE);
    LOG.debug("Exiting heartbeatWorker thread");
    return ret;
}




/**
 * Thread used to check if a response arrived in ctpCmdTimeout seconds.
 * If not, the CTP connection will be pulled down so that ctpThread
 * will restore the whole CTP connection.
 * This thread is started every time a message is sent.
 * @param lpv  unused
 */
DWORD WINAPI cmdTimeoutWorker(LPVOID lpv) {
    LOG.debug("Starting cmdTimeoutWorker thread");

    // Load the timeout (ctpCmdTimeout)
    CTPManager* ctpManager = CTPManager::getInstance();
    DWORD timeout = ctpManager->getConfig()->getCtpCmdTimeout();
    if (!timeout) {
        timeout = 180;      // 3 minutes max
    }

    Sleep(timeout * 1000);

    if ( (ctpManager->isLeaving() == false) &&
         (ctpManager->getCtpState() == CTP_STATE_WAITING_RESPONSE) ) {
        // Response not received -> close ctp connection so that
        // the receiveThread will exit with error, so ctpThread will restore ctp.
        LOG.info("No response received from Server after %d seconds: closing CTP", timeout);
        ctpManager->closeConnection();
    }

    LOG.debug("Exiting cmdTimeoutWorker thread");
    return 0;
}
